"""
Groups

This module contains actions for managing groups in Office 365.

!!! warning "Deprecated Plugin"

    This plugin will be retired in the future, please use the Azure AD plugin instead.

!!! warning "Mail-Enabled Groups"

    Unfortunately the Microsoft Graph API does not support creating or updating mail-enabled groups.

    These groups can be audited & added to others, but they cannot be created or updated using this plugin.

    See [here](https://docs.microsoft.com/en-us/graph/api/resources/groups-overview?view=graph-rest-1.0#security-groups-and-mail-enabled-security-groups) for more information.
"""
import logging
import functools

from typing import List, Union, Optional


from ppa import ui
from playbook_tools import tables, process


from . import _client, exceptions
from .users import _process_user
from .records import GraphClient, Group, GroupData, User


logger = logging.getLogger(__name__)


GROUP_TABLE_HEADER = ["Display Name", "Mail Nickname", "Mail Group", "Security Group", "Type(s)"]


def _row_generator(group) -> List[str]:
    return [
        group["display_name"],
        group["mail_nickname"] or "",
        str(group["mail_enabled"]),
        str(group["security_enabled"]),
        ", ".join(group["group_types"]) if group["group_types"] else "",
    ]


def _process_group(data) -> Group:
    return GroupData(**data).__dict__


@_client.graph_api
def get_all(client: GraphClient) -> List[Group]:
    """Get all groups.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __client:__ a [GraphClient](../inputs_outputs/#graphclient)

    ??? success "Output"

         A list of [Groups](../inputs_outputs/#group).

    ??? example
        ```YAML linenums="1"
        - microsoft_graph.groups.get_all:
          load:
            client: graph_secrets
          save: groups
        ```
    """
    logger.info("getting all groups")
    with _client._exception_handler("get groups"):
        return list(map(_process_group, _client.get(client, "/groups")))


@_client.graph_api
def get_by_id(client: GraphClient, group_id: str) -> Group:
    """Get the group with the supplied group ID.

    !!! question "Minimum Plugin Version: 1.1.0"

    ??? info "Input"

        - __client:__ a [GraphClient](../inputs_outputs/#graphclient)

        - __group_id__: the [Group](../inputs_outputs/#group) ID

    ??? success "Output"

         A single [Group](../inputs_outputs/#group).

    ??? example
        ```YAML linenums="1"
        - microsoft_graph.groups.get_by_id:
          load:
            group_id: group.id
            client: graph_secrets
          save: group
        ```
    """
    logger.info(f"getting group with ID '{group_id}'")
    with _client._exception_handler(f"get group with ID '{group_id}'"):
        return _process_group(_client.get(client, f"/groups/{group_id}"))


@_client.graph_api
def get_by_display_name(client: GraphClient, display_name: str) -> List[Group]:
    """Get all groups with the supplied display name.

    !!! question "Minimum Plugin Version: 1.1.0"

    !!! tip "Group display names are not unique!"

    ??? info "Input"

        - __client:__ a [GraphClient](../inputs_outputs/#graphclient)

        - __display_name__: the [Group](../inputs_outputs/#group) display name

    ??? success "Output"

        A [SearchResult](../../../searches/#search-result-dictionary) containing:

        - **total**: the number of [Groups](../inputs_outputs/#group) found

        - **all**: a list of [Groups](../inputs_outputs/#group) found

        - **first**: the first [Group](../inputs_outputs/#group) found

        - **last**: the last [Group](../inputs_outputs/#group) found

    ??? example
        ```YAML linenums="1"
        - microsoft_graph.groups.get_by_display_name:
          load:
            display_name: group.display_name
            client: graph_secrets
          save: search_result
        ```
    """
    logger.info(f"getting groups with display name '{display_name}'")
    with _client._exception_handler(f"get groups with display name '{display_name}'"):
        # The 'eq' filter doesn't work, so I'm using startsWith instead :(
        results = list(map(_process_group, _client.get(client, f"/groups?$filter=startsWith(displayName, '{display_name}')")))
    # Now do some client-side filtering to find only those groups whose display name *matches* what was supplied.
    return process.search_result([g for g in results if g["display_name"] == display_name])


@_client.graph_api
def get_users(client: GraphClient, group_id: str) -> List[User]:
    """Get users directly in the supplied group.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __client:__ a [GraphClient](../inputs_outputs/#graphclient)

        - __group_id:__ a [Group](../inputs_outputs/#group) ID

    ??? success "Output"

         A list of [Users](../inputs_outputs/#user).

    ??? example
        ```YAML linenums="1"
        - microsoft_graph.groups.get_users:
          load:
            group_id: selected_group.id
            client: graph_secrets
          save: users
        ```
    """
    logger.info(f"getting users directly in group with ID '{group_id}'")
    with _client._exception_handler("get users from group"):
        return list(map(_process_user, _client.get(client, f"/groups/{group_id}/members/microsoft.graph.user")))


@_client.graph_api
def get_groups(client: GraphClient, group_id: str) -> List[User]:
    """Get groups directly in the supplied group.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __client:__ a [GraphClient](../inputs_outputs/#graphclient)

        - __group_id:__ a [Group](../inputs_outputs/#group) ID

    ??? success "Output"

         A list of [Groups](../inputs_outputs/#group).

    ??? example
        ```YAML linenums="1"
        - microsoft_graph.groups.get_groups:
          load:
            group_id: selected_group.id
            client: graph_secrets
          save: groups
        ```
    """
    logger.info(f"getting groups directly in group with ID '{group_id}'")
    with _client._exception_handler("get groups from group"):
        return list(map(_process_group, _client.get(client, f"/groups/{group_id}/members/microsoft.graph.group")))


@_client.graph_api
def create_security_group(
    client: GraphClient,
    display_name: str,
    mail_nickname: str,
    owners: Optional[List[str]] = None,
    users: Optional[List[str]] = None,
    groups: Optional[List[str]] = None,
) -> None:
    """Create a security group.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __client:__ a [GraphClient](../inputs_outputs/#graphclient)

        - __display_name:__ a display name for the group

        - __mail_nickname:__ a mail alias for the group

        - __owners:__ optional list of [User](../inputs_outputs/#user) IDs if setting owners (defaults to no owners)

        - __users:__ optional list of [User](../inputs_outputs/#user) IDs to add as members

        - __groups:__ optional list of [Group](../inputs_outputs/#group) IDs to add as members

    !!! warning "Unified Groups"

        You cannot add Unified groups to Security groups.

    ??? success "Output"

         Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - microsoft_graph.groups.create:
            display_name: Example Group
            mail_nickname: example.group
          load:
            client: graph_secrets
        ```
    """
    owners, users, groups = [value if value is not None else [] for value in [owners, users, groups]]

    def _generate_ref(object_type, object_id):
        return "https://graph.microsoft.com/v1.0/{object_type}/{object_id}".format(object_type=object_type, object_id=object_id)

    post_data = {
        "displayName": display_name,
        "mailNickname": mail_nickname,
        "mailEnabled": False,
        "securityEnabled": True,
        "groupTypes": [],
    }

    # You can't supply None or an empty list in the post data.
    if owners := list(map(functools.partial(_generate_ref, "users"), owners)):
        post_data["owners@odata.bind"] = owners
    if members := list(map(functools.partial(_generate_ref, "users"), users)) + list(
        map(functools.partial(_generate_ref, "groups"), groups)
    ):
        post_data["members@odata.bind"] = members

    try:
        with _client._exception_handler(f"create group '{display_name}'"):
            return _client.post(client, "/groups", json=post_data)
    except exceptions.RequestError as e:
        if "a unified group cannot be added as a member" in str(e).lower():
            message = "Unified groups cannot be added to Security Groups."
            ui.output_error(message)
            raise exceptions.InvalidParameter(message)
        raise e


@_client.graph_api
def delete(client: GraphClient, group_id: str) -> None:
    """Delete a group.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __client:__ a [GraphClient](../inputs_outputs/#graphclient)

        - __group_id:__ a [Group](../inputs_outputs/#group) ID

    ??? success "Output"

         Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - microsoft_graph.groups.delete:
          load:
            group_id: selected_group.id
            client: graph_secrets
        ```
    """
    logger.info(f"deleting group with ID '{group_id}'")
    with _client._exception_handler("delete group"):
        return _client.delete(client, f"/groups/{group_id}")


@_client.graph_api
def add_member(client: GraphClient, group_id: str, member_id: str) -> None:
    """Add a member to a group.

    Members can be [Users](../inputs_outputs/#user) or [Groups](../inputs_outputs/#group).

    If the member is already in the group, nothing will be modified & the action will succeed.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __client:__ a [GraphClient](../inputs_outputs/#graphclient)

        - __group_id:__ a [Group](../inputs_outputs/#group) ID

        - __member_id:__ a [User](../inputs_outputs/#user) or [Group](../inputs_outputs/#group) ID

    !!! warning "Supported Group & Member Types"

        You cannot add Unified groups to Security groups.

        The Microsoft Graph API does not support adding users to:

        - Mail-Enabled Security Groups

        - Distribution Groups

    ??? success "Output"

         Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - microsoft_graph.groups.add_member:
          load:
            group_id: selected_group.id
            member_id: selected_user.id
            client: graph_secrets
        ```
    """
    logger.info(f"adding member with ID '{member_id}' to group with ID '{group_id}'")
    try:
        with _client._exception_handler("add group member", ignored_exceptions=[exceptions.AlreadyExists]):
            return _client.post(
                client,
                f"/groups/{group_id}/members/$ref",
                json={"@odata.id": f"https://graph.microsoft.com/v1.0/directoryObjects/{member_id}"},
            )
    except exceptions.RequestError as e:
        error = str(e).lower()
        if "a unified group cannot be added as a member" in error:
            message = "Unified groups cannot be added to Security Groups."
            ui.output_error(message)
            raise exceptions.InvalidParameter(message)
        elif "cannot update a mail-enabled security groups and or distribution list" in error:
            message = (
                "Adding users to distribution groups or mail-enabled security " "groups is not supported by the Microsoft API."
            )
            ui.output_error(message)
            raise exceptions.UnsupportedOperation(message)
        raise e


@_client.graph_api
def remove_member(client: GraphClient, group_id: str, member_id: str) -> None:
    """Remove a member from a group.

    Members can be [Users](../inputs_outputs/#user) or [Groups](../inputs_outputs/#group).

    If the member is not already in the group, nothing will be modified & the action will succeed.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __client:__ a [GraphClient](../inputs_outputs/#graphclient)

        - __group_id:__ a [Group](../inputs_outputs/#group) ID

        - __member_id:__ a [User](../inputs_outputs/#user) or [Group](../inputs_outputs/#group) ID

    ??? success "Output"

         Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - microsoft_graph.groups.remove_member:
          load:
            group_id: selected_group.id
            member_id: selected_user.id
            client: graph_secrets
        ```
    """
    logger.info(f"removing member with ID '{member_id}' from group with ID '{group_id}'")
    with _client._exception_handler("remove group member", ignored_exceptions=[exceptions.DoesNotExist]):
        return _client.delete(client, f"/groups/{group_id}/members/{member_id}/$ref")


def display(text: str, groups: Union[List[Group], Group]) -> None:
    """Display groups in a table.

The table will have the following columns:

- Display Name
- Mail Nickname
- Mail Enabled
- Security Enabled
- Type(s)

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __groups:__ any number of [Groups](../inputs_outputs/#group)

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - microsoft_graph.groups.get_all:
      load:
        client: graph_secrets
      save: groups

    - microsoft_graph.groups.display:
        text: All Groups
      load:
        groups: groups
    ```
    """
    tables.output_table(text, header=GROUP_TABLE_HEADER, extract=_row_generator, records=groups)


def select(text: str, groups: Union[List[Group], Group], minimum: Optional[int] = None, maximum: Optional[int] = None) -> None:
    """Display groups in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Display Name
- Mail Nickname
- Mail Enabled
- Security Enabled
- Type(s)

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __groups:__ any number of [Groups](../inputs_outputs/#group)

    - __minimum:__ the minimum number of selections

    - __maximum:__ the maximum number of selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [Groups](../inputs_outputs/#group)

    - **all**: a list of selected [Groups](../inputs_outputs/#group)

    - **first**: the first selected [Groups](../inputs_outputs/#group)

??? example

    ```YAML linenums="1"
    - microsoft_graph.groups.get_all:
      load:
        client: graph_secrets
      save: groups

    - microsoft_graph.groups.select:
        text: Select Groups
      load:
        groups: groups
      save: selection
    ```
    """
    return tables.input_table(
        text, header=GROUP_TABLE_HEADER, extract=_row_generator, records=groups, minimum=minimum, maximum=maximum
    )


def select_one(text: str, groups: Union[List[Group], Group]) -> None:
    """Display groups in a table & prompt the task operator to select one.

The table will have the following columns:

- Display Name
- Mail Nickname
- Mail Enabled
- Security Enabled
- Type(s)

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __groups:__ any number of [Groups](../inputs_outputs/#group)

??? success "Output"

    A single [Group](../inputs_outputs/#group).

??? example

    ```YAML linenums="1"
    - microsoft_graph.groups.get_all:
      load:
        client: graph_secrets
      save: groups

    - microsoft_graph.groups.select_one:
        text: Select a Group
      load:
        groups: groups
      save: group
    ```
    """
    return select(text, groups=groups, minimum=1, maximum=1)["first"]
